/* SPDX-License-Identifier: GPL-2.0-only */
/*
 * Copyright (c) 2021-2023 Qualcomm Innovation Center, Inc. All rights reserved.
 * Copyright (c) 2012-2021, The Linux Foundation. All rights reserved.
 */

#ifndef _DP_PARSER_H_
#define _DP_PARSER_H_

#include <linux/sde_io_util.h>
#include <msm_drv.h>

#define DP_LABEL "MDSS DP DISPLAY"
#define AUX_CFG_LEN	10
#define DP_MAX_PIXEL_CLK_KHZ	675000
#define DP_MAX_LINK_CLK_KHZ	810000
#define MAX_DP_MST_STREAMS	2
#define MAX_DP_BOND_NUM		3

enum dp_pm_type {
	DP_CORE_PM,
	DP_CTRL_PM,
	DP_PHY_PM,
	DP_STREAM0_PM,
	DP_STREAM1_PM,
	DP_LINK_PM,
	DP_PLL_PM,
	DP_BOND_PM,
	DP_MAX_PM
};

static inline const char *dp_parser_pm_name(enum dp_pm_type module)
{
	switch (module) {
	case DP_CORE_PM:	return "DP_CORE_PM";
	case DP_CTRL_PM:	return "DP_CTRL_PM";
	case DP_PHY_PM:		return "DP_PHY_PM";
	case DP_STREAM0_PM:	return "DP_STREAM0_PM";
	case DP_STREAM1_PM:	return "DP_STREAM1_PM";
	case DP_LINK_PM:	return "DP_LINK_PM";
	case DP_PLL_PM:		return "DP_PLL_PM";
	case DP_BOND_PM:	return "DP_BOND_PM";
	default:		return "???";
	}
}

/**
 * struct dp_display_data  - display related device tree data.
 *
 * @ctrl_node: referece to controller device
 * @phy_node:  reference to phy device
 * @is_active: is the controller currently active
 * @name: name of the display
 * @display_type: type of the display
 */
struct dp_display_data {
	struct device_node *ctrl_node;
	struct device_node *phy_node;
	bool is_active;
	const char *name;
	const char *display_type;
};

/**
 * struct dp_io_data - data structure to store DP IO related info
 * @name: name of the IO
 * @buf: buffer corresponding to IO for debugging
 * @io: io data which give len and mapped address
 */
struct dp_io_data {
	const char *name;
	u8 *buf;
	struct dss_io_data io;
};

/**
 * struct dp_io - data struct to store array of DP IO info
 * @len: total number of IOs
 * @data: pointer to an array of DP IO data structures.
 */
struct dp_io {
	u32 len;
	struct dp_io_data *data;
};

/**
 * struct dp_pinctrl - DP's pin control
 *
 * @pin: pin-controller's instance
 * @state_active: active state pin control
 * @state_hpd_active: hpd active state pin control
 * @state_suspend: suspend state pin control
 */
struct dp_pinctrl {
	struct pinctrl *pin;
	struct pinctrl_state *state_active;
	struct pinctrl_state *state_hpd_active;
	struct pinctrl_state *state_hpd_tlmm;
	struct pinctrl_state *state_hpd_ctrl;
	struct pinctrl_state *state_suspend;
};

#define DP_ENUM_STR(x)	#x
#define DP_AUX_CFG_MAX_VALUE_CNT 3
/**
 * struct dp_aux_cfg - DP's AUX configuration settings
 *
 * @cfg_cnt: count of the configurable settings for the AUX register
 * @current_index: current index of the AUX config lut
 * @offset: register offset of the AUX config register
 * @lut: look up table for the AUX config values for this register
 */
struct dp_aux_cfg {
	u32 cfg_cnt;
	u32 current_index;
	u32 offset;
	u32 lut[DP_AUX_CFG_MAX_VALUE_CNT];
};

/**
 * struct dp_msa_param - DP panel's MSA configuration
 *
 * @ovr_visible_width_in_px:  width of the image in pixels
 * @ovr_visible_height_in_px: height of the image in pixels
 * @ovr_h_back_porch_px: horizontal back porch of the image in pixels
 * @ovr_h_front_porch_px: horizontal front porch of the image in pixels
 * @ovr_h_sync_pulse_px: horizontal sync pulse width in pixels
 * @ovr_h_sync_skew_px: horizontal sync skew in pixels
 * @ovr_v_back_porch_ln: vertical back porch of the image in lines
 * @ovr_v_front_porch_ln: vertical front porch of the image in lines
 * @ovr_v_sync_pulse_ln: vertical sync pulse width in lines
 * @ovr_h_left_border_px: horizontal left border in pixels
 * @ovr_h_right_border_px: horizontal right border in pixels
 * @ovr_v_top_border_ln: vertical top border in lines
 * @ovr_v_bottom_border_ln: vertical bottom border in lines
 * @ovr_h_sync_active_low: set to 1 if horizontal sync pulse is active low
 * @ovr_v_sync_active_low: set to 1 if vertical sync pulse is active low
 * @ovr_sw_mvid: MVID parameters used by sync device
 * @ovr_sw_nvid: NVID parameter used by sync device
 * @ovr_v_refresh_rate: vertical refresh rate in Hz.
 */
struct dp_msa_param {
	u32 ovr_visible_width_in_px;
	u32 ovr_visible_height_in_px;
	u32 ovr_h_back_porch_px;
	u32 ovr_h_front_porch_px;
	u32 ovr_h_sync_pulse_px;
	u32 ovr_h_sync_skew_px;
	u32 ovr_v_back_porch_ln;
	u32 ovr_v_front_porch_ln;
	u32 ovr_v_sync_pulse_ln;
	u32 ovr_h_left_border_px;
	u32 ovr_h_right_border_px;
	u32 ovr_v_top_border_ln;
	u32 ovr_v_bottom_border_ln;
	u32 ovr_h_sync_active_low;
	u32 ovr_v_sync_active_low;
	u32 ovr_sw_mvid;
	u32 ovr_sw_nvid;
	u32 ovr_v_refresh_rate;
};
/* PHY AUX config registers */
enum dp_phy_aux_config_type {
	PHY_AUX_CFG0,
	PHY_AUX_CFG1,
	PHY_AUX_CFG2,
	PHY_AUX_CFG3,
	PHY_AUX_CFG4,
	PHY_AUX_CFG5,
	PHY_AUX_CFG6,
	PHY_AUX_CFG7,
	PHY_AUX_CFG8,
	PHY_AUX_CFG9,
	PHY_AUX_CFG_MAX,
};

/**
 * enum dp_phy_version - version of the dp phy
 * @DP_PHY_VERSION_UNKNOWN: Unknown controller version
 * @DP_PHY_VERSION_4_2_0:   DP phy v4.2.0 controller
 * @DP_PHY_VERSION_5_0_0:   DP phy v5.0.0 controller
 * @DP_PHY_VERSION_6_0_0:   DP phy v6.0.0 controller
 * @DP_PHY_VERSION_MAX:     max version
 */
enum dp_phy_version {
	DP_PHY_VERSION_UNKNOWN,
	DP_PHY_VERSION_2_0_0 = 0x200,
	DP_PHY_VERSION_4_2_0 = 0x420,
	DP_PHY_VERSION_5_0_0 = 0x500,
	DP_PHY_VERSION_6_0_0 = 0x600,
	DP_PHY_VERSION_MAX
};


/**
 * enum dp_phy_bond_mode - bond mode of the dp phy
 * @DP_PHY_BOND_MODE_NONE:        Non-bond mode
 * @DP_PHY_BOND_MODE_PLL_MASTER:  PLL bond mode master
 * @DP_PHY_BOND_MODE_PLL_SLAVE:   PLL bond mode slave
 * @DP_PHY_BOND_MODE_PCLK_MASTER: Pixel clock bond mode master
 * @DP_PHY_BOND_MODE_PCLK_SLAVE:  Pixel clock bond mode slave
 * @DP_PHY_BOND_MODE_MAX:         max bond mode
 */
enum dp_phy_bond_mode {
	DP_PHY_BOND_MODE_NONE = 0,
	DP_PHY_BOND_MODE_PLL_MASTER,
	DP_PHY_BOND_MODE_PLL_SLAVE,
	DP_PHY_BOND_MODE_PCLK_MASTER,
	DP_PHY_BOND_MODE_PCLK_SLAVE,
	DP_PHY_BOND_MODE_MAX
};

#define IS_BOND_MODE(x)		((x) != DP_PHY_BOND_MODE_NONE)
#define IS_PLL_BOND_MODE(x)	\
	((x) == DP_PHY_BOND_MODE_PLL_MASTER \
		|| (x) == DP_PHY_BOND_MODE_PLL_SLAVE)
#define IS_PCLK_BOND_MODE(x)	\
	((x) == DP_PHY_BOND_MODE_PCLK_MASTER \
		|| (x) == DP_PHY_BOND_MODE_PCLK_SLAVE)
#define IS_BOND_MODE_MASTER_PHY(x)	\
	((x) == DP_PHY_BOND_MODE_PLL_MASTER \
		|| (x) == DP_PHY_BOND_MODE_PCLK_MASTER)
#define IS_BOND_MODE_SLAVE_PHY(x)	\
	((x) == DP_PHY_BOND_MODE_PLL_SLAVE \
		|| (x) == DP_PHY_BOND_MODE_PCLK_SLAVE)

/**
 * enum dp_phy_mode - mode of the dp phy
 * @DP_PHY_MODE_UNKNOWN: Unknown PHY mode
 * @DP_PHY_MODE_DP:      DP PHY mode
 * @DP_PHY_MODE_MINIDP:  MiniDP PHY mode
 * @DP_PHY_MODE_EDP:     eDP PHY mode
 * @DP_PHY_MODE_EDP_HIGH_SWING:   eDP PHY mode, high swing/pre-empahsis
 * @DP_PHY_MODE_MAX:     max PHY mode
 */
enum dp_phy_mode {
	DP_PHY_MODE_UNKNOWN = 0,
	DP_PHY_MODE_DP,
	DP_PHY_MODE_MINIDP,
	DP_PHY_MODE_EDP,
	DP_PHY_MODE_EDP_HIGH_SWING,
	DP_PHY_MODE_MAX
};

/**
 * struct dp_hw_cfg - DP HW specific configuration
 *
 * @phy_version: DP PHY HW version
 */
struct dp_hw_cfg {
	enum dp_phy_version phy_version;
	enum dp_phy_mode phy_mode;
};

enum dp_bond_type {
	DP_BOND_DUAL_PHY,
	DP_BOND_DUAL_PCLK,
	DP_BOND_TRIPLE_PHY,
	DP_BOND_TRIPLE_PCLK,
	DP_BOND_MAX,
};

static const int num_bond_dp[DP_BOND_MAX] = { 2, 2, 3, 3 };

struct dp_bond_cfg {
	bool enable;
	u32 ctrl[MAX_DP_BOND_NUM];
};

static inline char *dp_phy_aux_config_type_to_string(u32 cfg_type)
{
	switch (cfg_type) {
	case PHY_AUX_CFG0:
		return DP_ENUM_STR(PHY_AUX_CFG0);
	case PHY_AUX_CFG1:
		return DP_ENUM_STR(PHY_AUX_CFG1);
	case PHY_AUX_CFG2:
		return DP_ENUM_STR(PHY_AUX_CFG2);
	case PHY_AUX_CFG3:
		return DP_ENUM_STR(PHY_AUX_CFG3);
	case PHY_AUX_CFG4:
		return DP_ENUM_STR(PHY_AUX_CFG4);
	case PHY_AUX_CFG5:
		return DP_ENUM_STR(PHY_AUX_CFG5);
	case PHY_AUX_CFG6:
		return DP_ENUM_STR(PHY_AUX_CFG6);
	case PHY_AUX_CFG7:
		return DP_ENUM_STR(PHY_AUX_CFG7);
	case PHY_AUX_CFG8:
		return DP_ENUM_STR(PHY_AUX_CFG8);
	case PHY_AUX_CFG9:
		return DP_ENUM_STR(PHY_AUX_CFG9);
	default:
		return "unknown";
	}
}

/**
 * struct dp_dsc_passthrough - data struct to store DSC pass through related data
 *
 * @dsc_passthrough_enable: Enables DSC forcefully even if panel doesn't support it.
 * @dsc_force_fec_enable: Enables FEC forcefully.
 * @comp_info: DSC compression info including PPS.
 */
struct dp_dsc_passthrough {
	bool dsc_passthrough_enable;
	struct msm_compression_info comp_info;
};

/**
 * struct dp_parser - DP parser's data exposed to clients
 *
 * @pdev: platform data of the client
 * @mp: gpio, regulator and clock related data
 * @pinctrl: pin-control related data
 * @disp_data: controller's display related data
 * @cell_idx: index of the display
 * @l_pnswap: P/N swap status on each lane
 * @max_pclk_khz: maximum pixel clock supported for the platform
 * @max_lclk_khz: maximum link clock supported for the platform
 * @hw_cfg: DP HW specific settings
 * @has_mst: MST feature enable status
 * @has_mst_sideband: MST sideband feature enable status
 * @no_aux_switch: presence AUX switch status
 * @gpio_aux_switch: presence GPIO AUX switch status
 * @dsc_feature_enable: DSC feature enable status
 * @fec_feature_enable: FEC feature enable status
 * @dsc_continuous_pps: PPS sent every frame by HW
 * @has_widebus: widebus (2PPC) feature eanble status
 * @force_bond_mode: force dp in bond mode
 * @force_connect_mode: force dp in connect mode
 * @no_link_rate_reduction: skip link rate reduction during link training
 * @no_lane_count_reduction: skip lane count reduction during link training
 * @force_connect_mode: force dp in connect mode
 * @mst_fixed_port: mst port_num reserved for fixed topology
 * @mst_fixed_display_type: mst display_type reserved for fixed topology
 * @display_type: display type as defined in device tree.
 * @qos_cpu_mask: CPU mask for QOS
 * @qos_cpu_latency: CPU Latency setting for QOS
 * @parse: function to be called by client to parse device tree.
 * @get_io: function to be called by client to get io data.
 * @get_io_buf: function to be called by client to get io buffers.
 * @clear_io_buf: function to be called by client to clear io buffers.
 */
struct dp_parser {
	struct platform_device *pdev;
	struct dss_module_power mp[DP_MAX_PM];
	struct dp_pinctrl pinctrl;
	struct dp_io io;
	struct dp_display_data disp_data;
	u32 cell_idx;
	struct dp_bond_cfg bond_cfg[DP_BOND_MAX];
	struct dp_dsc_passthrough dsc_passthrough;
	struct dp_msa_param msa;

	u8 l_map[4];
	u8 l_pnswap;
	struct dp_aux_cfg aux_cfg[AUX_CFG_LEN];
	u32 max_pclk_khz;
	u32 max_lclk_khz;
	struct dp_hw_cfg hw_cfg;
	bool has_mst;
	bool has_mst_sideband;
	bool no_aux_switch;
	bool dsc_feature_enable;
	bool msa_config;
	bool fec_feature_enable;
	bool dsc_continuous_pps;
	bool has_widebus;
	bool has_force_encryption;
	bool hdcp_wait_sink_sync_enabled;
	bool gpio_aux_switch;
	bool lphw_hpd;
	bool force_bond_mode;
	bool force_connect_mode;
	bool no_link_rate_reduction;
	bool no_lane_count_reduction;
	u32 link_training_min_vlevel;
	u32 link_training_min_plevel;
	u32 mst_fixed_port[MAX_DP_MST_STREAMS];
	u32 pixel_base_off[MAX_DP_MST_STREAMS];
	u32 qos_cpu_mask;
	unsigned long qos_cpu_latency;
	const char *mst_fixed_display_type[MAX_DP_MST_STREAMS];
	const char *display_type;
	u32 aux_timeout;
	u32 aux_retry_count;

	int (*parse)(struct dp_parser *parser);
	struct dp_io_data *(*get_io)(struct dp_parser *parser, char *name);
	void (*get_io_buf)(struct dp_parser *parser, char *name);
	void (*clear_io_buf)(struct dp_parser *parser);
};

enum dp_phy_lane_num {
	DP_PHY_LN0 = 0,
	DP_PHY_LN1 = 1,
	DP_PHY_LN2 = 2,
	DP_PHY_LN3 = 3,
	DP_MAX_PHY_LN = 4,
};

enum dp_mainlink_lane_num {
	DP_ML0 = 0,
	DP_ML1 = 1,
	DP_ML2 = 2,
	DP_ML3 = 3,
};

/**
 * dp_parser_get() - get the DP's device tree parser module
 *
 * @pdev: platform data of the client
 * @cell_idx: index of the DP display
 * return: pointer to dp_parser structure.
 *
 * This function provides client capability to parse the
 * device tree and populate the data structures. The data
 * related to clock, regulators, pin-control and other
 * can be parsed using this module.
 */
struct dp_parser *dp_parser_get(struct platform_device *pdev, u32 cell_idx);

/**
 * dp_parser_put() - cleans the dp_parser module
 *
 * @parser: pointer to the parser's data.
 */
void dp_parser_put(struct dp_parser *parser);
#endif
